/**
 * $Id: main.c,v 1.3 2009/03/12 15:00:51 thomasw Exp $
 *
 * Copyright (C) 2007 Sealevel Systems Inc.
 *   
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Sealevel Systems Inc.
 * 155 Technology Place
 * P.O. Box 830
 * Liberty, SC 29657
 * www.sealevel.com
 * 
 *
 * For support please contact us by phone at (864) 843-4343 during normal
 * bussiness hours (8am - 5pm EST, M-F) or e-mail us at support@sealevel.com
 *
 *
 * This is a very simple application to enable shared interrupts on
 * Supported PCI Express Sealevel Systems async cards.  This is accomplished
 * by enabling bit 11 in register 0x68 of the PCI memory region.  If you desire,
 * there is also code showing how do disable the shared interrupts as well
 * (effectively this is the same as deactivating the card...)
 *
 * The PCI memory region is accessed using mmap() on /dev/mem.  For now this
 * seems like the way to go for the most compatability.  It may eventually
 * become necessary (and easier) to access through the /sys fs.
 *
 * Use of this code requires the PCI-Utils-dev package version 2.2.4 or
 * greater.  This package is available with most distributions or can be
 * obtained directly from the author at http://mj.ucw.cz/pciutils.shtml
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <pci/pci.h>
#include <sys/io.h>
#include <sys/mman.h>
#include <string.h>

#define OUT(mess, args...) fprintf(stdout, mess "\n", ##args)
#define ERR(mess, args...) fprintf(stderr,"INTENABLE ERR: " mess "\n", ##args)

// You can enable verbose debugging output
#define IFIDEBUG(mess, args...) if(debug) OUT(" **DEBUG: " mess, ##args) 

#define VENDOR 0x135E  //Sealevel Systems Vendor ID
#define DEVICE 0xE161  //The PCI Express 16 port device ID
#define DEVICE2 0xE205 //The PCI Express 2 port device ID
// TODO: Create device ID table and search function to simplify conditional

static int debug = 0;
static int enable = 1;
static const char *usage = "Enable or Disable shared interrupts on Sealevel\n"
			   "Systems, Inc. PCI Express async serial devices.\n"
			   "\n"
			   "\n"
			   "Options\n"
			   "  help, -help, --help  Display usage & exit.\n"
			   "  -v, --version        Display version & exit.\n"
			   "  -b, --debug          Enable debug mode.\n"
			   "  -e, --enable         Enable shared interrupts.\n"
			   "  -d, --disable        Disable shared interrupts.\n"
			   "\n"
			   "Default behavior is to enable shared interrupts.\n";
			   
static const char *version = "VERSION: $Name:  $:$Date: 2009/03/12 15:00:51 $\n";

// ----------------------------------------------------------------------------
// No args necessary.  All it does it look for any and all PCI cards with
// vendor ID of VENDOR and device ID of DEVICE.  And set bit 11 of register 
// 0x68 in the first memory region (BAR 0).  But you may specify debug level.
// ----------------------------------------------------------------------------
int main( int argc, char *argv[])
{ 
	struct pci_access *all_devices;          // pointer to all PCI devs
	struct pci_dev *dev_cursor;              // pointer to a single PCI dev

	unsigned int page_size;                  // size of a page file  
	u32 pci_mem_addr;                        // system address of pci_mem
	char *pci_mem = 0x00;                    // mapped location of pci_mem

	int fd;                                  // file descriptor (/dev/mem)
	int i;					 // counter (for loop)

	// Completely fluff
	for( i = 1; i < argc; i++ )
	{
		// Display usage and exit
		if( strcmp("help",argv[i]) == 0 || strcmp("-help",argv[i]) == 0
		    || strcmp("--help",argv[i]) == 0)
		{ 
			OUT("%s", usage);
			OUT("%s", version);
			return 0;
		}
		
		// Display version and exit
		if(strcmp("-v",argv[i]) == 0 || strcmp("--version",argv[i]) ==0)
		{
			OUT("%s", version);
			return 0;
		}
		
		// Set debug
		if(strcmp("-b",argv[i]) == 0 || strcmp("--debug",argv[i]) == 0)
			debug = 1;
		
		// set enable true
		if(strcmp("-e",argv[i]) == 0 || strcmp("--enable",argv[i]) == 0)
			enable = 1;
		
		// set enable false
		if(strcmp("-d",argv[i]) == 0 || strcmp("--disable",argv[i]) ==0)
			enable = 0;
	}

	IFIDEBUG("%s", version);
	IFIDEBUG("debug = %d", debug);
	IFIDEBUG("enable = %d", enable);

	// Check IO permissions to be able to open /dev/mem
	if(iopl(3))
	{
		ERR("Cannot access /dev/mem (try re-running as root)");
		return -1;
	}
	
	// Go ahead and open physical memory special file
	fd = open ( "/dev/mem", O_RDWR);
	if (fd < 0)
	{
		ERR("Failed to open /dev/mem");
		return -1;
	}

	// scan the PCI bus
	all_devices = pci_alloc();            // Allocate memory for all_devices
	pci_init(all_devices);                // Initialisation of all_devices
	pci_scan_bus(all_devices);            // Scan the PCI bus(es)

	// Get the size of a page (Because we have to map a whole page)
	page_size = getpagesize();

	// cycle through the devices looking for any of ours
	dev_cursor = all_devices->devices;
	while (dev_cursor != NULL)
	{
		unsigned short ven = dev_cursor->vendor_id;
		unsigned short dev = dev_cursor->device_id;
		
		IFIDEBUG("ven= %04X, dev= %04X", ven, dev);
		
		if( ven == VENDOR && (dev == DEVICE || dev == DEVICE2) )
		{
			unsigned char b = dev_cursor->bus;
			unsigned char d = dev_cursor->dev;
			unsigned char f = dev_cursor->func;
			unsigned int mask = page_size - 1;
			
			OUT("%04X:%04X found at %02X:%02X.%X", ven,dev,b,d,f);
		
			// BAR 0 (0x10) is the offset for the first region 512
			pci_mem_addr = pci_read_long(dev_cursor, 0x10);
			pci_mem_addr &= PCI_BASE_ADDRESS_MEM_MASK;
			
			IFIDEBUG("physical pci mem addr = %08X", pci_mem_addr);
			IFIDEBUG("page_size = %d, mask = %08X", page_size,mask);
			
			// As mentioned, since 512 is so small, we can't just
			// map the pci memory region, have to map a whole page
			// ~mask clears the bits containing the local address
			// of the 512 byte segment within the larger page
			// ie.  ps = 1024 (0x400), mask = 1023 (0x3FF),
			// phys addr = 1536 (0x600), ~mask = 3072 (0xC00),
			// map = phys & ~mask = (0x600)&(0xC00) = 1024 (0x600)
			// so we get 1024 bytes starting at byte 1024.
			//
			// To get to our 512 bytes in this page, we can use mask
			// phys addr = 1536 (0x600), mask = 1023 (0x3FF),
			// start of pci_mem = phys & mask = (0x600)&(0x3FF) =
			// 512 (0x200), which we already knew.
			pci_mem = (char *) mmap(NULL, page_size, 
						PROT_READ|PROT_WRITE, 
						MAP_SHARED|MAP_FIXED, 
						fd, 
						(off_t)(pci_mem_addr & ~mask)
						);
						
			if (pci_mem == MAP_FAILED)
			{
				ERR("Failed to map pci memory");
				close(fd);
				return -1;
			}

			IFIDEBUG("mapped page = %02X", (unsigned int) pci_mem);
			IFIDEBUG("pci_mem mask = %08X", (pci_mem_addr & mask));
			
			// Bit 11 must be set in this 32-bit register
			IFIDEBUG("long reg 0x68= 0x%02X %02X %02X %02X", 
				 pci_mem[ (pci_mem_addr & mask) + 0x6B ],
				 pci_mem[ (pci_mem_addr & mask) + 0x6A ], 
				 pci_mem[ (pci_mem_addr & mask) + 0x69 ],
				 pci_mem[ (pci_mem_addr & mask) + 0x68 ] );


			// Here is what we came to do
			if (enable)
			{
				pci_mem[ (pci_mem_addr & mask) + 0x69 ] |= 0x08;
				IFIDEBUG("setting bit 11: 0x XX XX %02X XX",
					pci_mem[ (pci_mem_addr & mask) + 0x69]);
			}
			else
			{
				pci_mem[ (pci_mem_addr & mask) + 0x69 ] &= 0xF3;
				IFIDEBUG("clearing bit 11: 0x XX XX %02X XX",
					pci_mem[ (pci_mem_addr & mask) + 0x69]);
			}

			IFIDEBUG("long reg 0x68= 0x%02X %02X %02X %02X", 
				 pci_mem[ (pci_mem_addr & mask) + 0x6B ],
				 pci_mem[ (pci_mem_addr & mask) + 0x6A ], 
				 pci_mem[ (pci_mem_addr & mask) + 0x69 ],
				 pci_mem[ (pci_mem_addr & mask) + 0x68 ] );			

			// Did it work?  I did this rather than error out
			// in case someone decides to disable the interrupts
			if( (pci_mem[ (pci_mem_addr & mask) + 0x69 ] & 0x08) )
				OUT("  Shared interrupts: \tENABLED");
			else
				OUT("  Shared interrupts: \tDISABLED");
		}

		dev_cursor = dev_cursor->next;  //next device

	}

	// Only here to force next line
	OUT("");

	// Clean up and exit
	munmap(pci_mem, page_size);
	close(fd);

	return 0;
}
