/**
 * $Id: main.c,v 1.2 2009/03/12 15:00:51 thomasw Exp $
 *
 * Copyright (C) 2008 Sealevel Systems Inc.
 *   
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Sealevel Systems Inc.
 * 155 Technology Place
 * P.O. Box 830
 * Liberty, SC 29657
 * www.sealevel.com
 * 
 *
 * For support please contact us by phone at (864) 843-4343 during normal
 * bussiness hours (8am - 5pm EST, M-F) or e-mail us at support@sealevel.com
 *
 * 
 * This utility is designed to allow the selection of electrical interface 
 * with Sealevel Systems, Inc. USB devices.  It is a very simple command
 * line tool to allow acquiring and chaning the current electrical interface
 * (rs mode) of a Sealevel software selectable USB device.
 *
 * This application requires use of the FTDI Direct Device library.  This
 * library interacts with FTDI devices through use of the USBFS.  If this
 * file system is not mounted in /proc/bus/usb, this utility will not work.
 * View the ftd2xx documentation in the /lib/shared directory of the Seacom
 * package available from www.sealevel.com
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "setusb/interface.h"
#include "setusb/slusb.h"
#include "ftd2xx/ftd2xx.h"

#define OUT(mess, args...) fprintf(stdout, mess "\n", ##args)
#define ERR(mess, args...) fprintf(stderr,"INTENABLE ERR: " mess "\n", ##args)

// You can enable verbose debugging output
int debug = 0;
#define IFIDEBUG(mess, args...) if(debug) OUT(" **DEBUG: " mess, ##args) 

int quiet = 0;

static const char *usage = "Get or set the async electrical interface\n"
			   "on Sealevel Systems, Inc. USB serial converters.\n"
			   "\n"
			   "\n"
			   "Options\n"
			   "  -h, -?, --help       Display usage & exit.\n"
			   "  -v, --version        Display version & exit.\n"
			   "  -b, --debug          Enable debug mode.\n"
			   "  -q, --quiet          Suppress output.\n"
			   "  -d, --device         Specify the index of the device.\n"
			   "  -p, --port           Specify the port number.\n"
			   "  -i, --interface      Specify the interface.\n"
			   "\n";
			   
static const char *version = "VERSION: $Name:  $:$Date: 2009/03/12 15:00:51 $\n";

int parseOptions(int argc, char *argv[], int *device, int *port, Interface *interface)
{
	int i;

	// Loop for arguments
	for(i = 0; i < argc; i++)
	{
		// Display usage and exit
		if(strcmp("-h", argv[i]) == 0 || strcmp("-?", argv[i]) == 0
		    || strcmp("--help", argv[i]) == 0)
		{ 
			OUT("%s", usage);
			OUT("%s", version);
			return -1;
		}
		
		// Display version and exit
		if(strcmp("-v", argv[i]) == 0 || strcmp("--version", argv[i]) == 0)
		{
			OUT("%s", version);
			return -1;
		}
	
		// Actual parameters
		// Set debug
		if(strcmp("-b", argv[i]) == 0 || strcmp("--debug", argv[i]) == 0)
		{
			debug = 1;
		}
		// Set quiet
		else if(strcmp("-q", argv[i]) == 0 || strcmp("--quiet", argv[i]) == 0)
		{
			quiet = 1;
		}
		// Specify device
		else if(strcmp("-d", argv[i]) == 0 || strcmp("--device", argv[i]) == 0)
		{
			i++;

			if (strcmp("ALL", argv[i]) == 0)
			{
				*device = -1;
			}
			else
			{
				*device = atoi(argv[i]);
			}
		}
		// Specify port
		else if (strcmp("-p", argv[i]) == 0 || strcmp("--port", argv[i]) == 0)
		{
			i++;

			if (strcmp("ALL", argv[i]) == 0)
			{
				*port = -1;
			}
			else
			{
				*port = atoi(argv[i]);
			}
		}
		// Specify interface selection
		else if (strcmp("-i", argv[i]) == 0 || strcmp("--interface", argv[i]) == 0)
		{
			i++;

			*interface = atointerface(argv[i]);
		}
		// Default = blow up, unknown argument
		else
		{
			OUT("Unknown option: %s", argv[i]);
			OUT("Try -h for usage information");
			return -1;
		}
	}

	// good to go
	return 0;
}

void printUSBSettings()
{
	int i;

	OUT("%i total Software Selectable Sealevel USB devices detected.", slusb_getNumDevices());

	// Print current interface settings of all supported USB devices
	for (i = 0; i < slusb_getNumDevices(); i++)
	{
		int j;

		DeviceConfig config = slusb_getInterfaces(i);

		OUT("%i:\t%i", i, config.model);
		for (j = 0; j < config.numports; j++)
		{
			OUT("\t%i\t%s", j, interfaceToString(config.interface[j]));
		}
		OUT("");
	}
}

int setInterfaces(int device, int port, Interface interface)
{
	DeviceConfig config = slusb_getInterfaces(device);

	if (port > config.numports)
	{
		ERR("Cannot set port %i on device %i... only has %i ports total\n", 
			port, device, config.numports);
	}

	if (port < 0)
	{
		int j;
		for (j = 0; j < config.numports; j++)
			config.interface[j] = interface;
	}
	else
	{
		config.interface[port] = interface;
	}

	return slusb_setInterfaces(device, config);
}

int setAllInterfaces(int port, Interface interface)
{
	int i;

	for (i = 0; i < slusb_getNumDevices(); i++)
	{
		setInterfaces(i, port, interface);
	}

	return 0;
}

// ----------------------------------------------------------------------------
int main(int argc, char *argv[])
{
	int ret;
	int device = -1;
	int port = -1;
	Interface interface = UNDEFINED;

	ret = parseOptions((argc-1), &argv[1], &device, &port, &interface);
	if (ret < 0)
	{
		return ret;
	}

	// Remove ftdi_sio if present
	system("modprobe -r ftdi_sio");

	// If interface is anything other than "Undefined", we are setting
	if (interface != UNDEFINED)
	{
		// Check for presence of all devices wildcard...
		if(device < 0)
		{
			setAllInterfaces(port, interface);
		}
		else
		{
			setInterfaces(device, port, interface);
		}
	}

	// Return the current settings of specified device/port
	if (!quiet)
	{
		printUSBSettings();
	}

	// Reinsert ftdi_sio
	system("modprobe ftdi_sio");

	return 0;
}
// ----------------------------------------------------------------------------

FT_HANDLE slusb_openDevice(int device, Device *info)
{
	int i;
	FT_HANDLE handle;
	DWORD oldVID, oldPID;

	if (device < 0 || device >= slusb_getNumDevices())
	{
		return 0;
	}

	// Find the FTDI device index
	// Cycle through each Device in our supported list and see how many there are...
	// when we get to the device which contains the one of interest... stop
	for (i = 0; i < sizeof(supportedUSBList)/sizeof(Device); i++)
	{
		int number = slusb_getNumDevicesByType(supportedUSBList[i]);

		if ((device - number) < 0)
		{
			info->vendor = supportedUSBList[i].vendor;
			info->model = supportedUSBList[i].model;
			info->numports = supportedUSBList[i].numports;
			info->name = supportedUSBList[i].name;
			break;
		}
		else
		{
			device -= number;
		}
	}

	FT_GetVIDPID(&oldVID, &oldPID);
	FT_SetVIDPID(supportedUSBList[i].vendor, supportedUSBList[i].model);
	// Remember each chip is dual channel, so multiply by 2!
	FT_Open((device * 2), &handle);
	FT_SetVIDPID(oldVID, oldPID);

	return handle;
}

// Return the number of currently plugged in devices total
int slusb_getNumDevices()
{
	int numDevices = 0;
	int i;
	
	for (i = 0; i < sizeof(supportedUSBList)/sizeof(Device); i++)
		numDevices += slusb_getNumDevicesByType(supportedUSBList[i]);

	return numDevices;
}

// Return the number of currently plugged in devices of type Device
int slusb_getNumDevicesByType(Device type)
{
	int numDevices = 0;
	DWORD oldVID, oldPID, total;

	// Save current custom IDs for restoration later
	FT_GetVIDPID(&oldVID, &oldPID);

	// Clear the custom so we can see how many total FTDI devices there are
	FT_SetVIDPID(0x0000, 0x0000);
	FT_ListDevices(&total, NULL, FT_LIST_NUMBER_ONLY);
	numDevices -= total;

	// Set our custom IDs so we can see how many FTDI devices + our custom devices there are
	FT_SetVIDPID(type.vendor, type.model);
	FT_ListDevices(&total, NULL, FT_LIST_NUMBER_ONLY);
	numDevices += total;

	FT_SetVIDPID(oldVID, oldPID);

	// Each chip is dual channel, so divide by two to get the actual number
	return (numDevices / 2);
}

// Get the currently active electrical interfaces for the specified device
// zero-based index
DeviceConfig slusb_getInterfaces(int device)
{
	int i;
	FT_HANDLE handle;
	Device info;
	DeviceConfig config;
	unsigned char *eepromdata;
	DWORD size, read;

	initDeviceConfig(&config);

	handle = slusb_openDevice(device, &info);
	if (handle <= 0)
	{
		return config;
	}

	// Device is now open and ready to read EEPROM and decode it
	if (FT_EE_UASize(handle, &size) != FT_OK)
	{
		FT_Close(handle);
		return config;
	}

	eepromdata = (unsigned char *) malloc(size);
	if (eepromdata == NULL)
	{
		FT_Close(handle);
		return config;
	}

	FT_EE_UARead(handle, eepromdata, size, &read);
	if (read < (10 + info.numports))
	{
		free(eepromdata);
		FT_Close(handle);
		return config;
	}

	// Now decode the interfaces part of the data
	for (i = 0; i < info.numports; i++)
	{
		config.interface[i] = convertToInterface(eepromdata[EEPROM_OFFSET + i]);
	}
	config.numports = info.numports;
	config.model = info.name;

	free(eepromdata);
	FT_Close(handle);

	return config;
}

// Apply a device config across an entire device
int slusb_setInterfaces(int device, DeviceConfig config)
{
	int i;
	FT_HANDLE handle;
	Device info;
	unsigned char *eepromdata;
	DWORD size, read;

	handle = slusb_openDevice(device, &info);
	if (handle <= 0)
	{
		return -1;
	}

	// Device is now open and ready to read EEPROM and write it
	if (FT_EE_UASize(handle, &size) != FT_OK)
	{
		FT_Close(handle);
		return -2;
	}

	eepromdata = (unsigned char *) malloc(size);
	if (eepromdata == NULL)
	{
		FT_Close(handle);
		return -3;
	}

	FT_EE_UARead(handle, eepromdata, size, &read);
	if (read < (10 + info.numports))
	{
		free(eepromdata);
		FT_Close(handle);
		return -4;
	}

	// Now overwrite the interfaces part of the data
	for (i = 0; i < info.numports; i++)
	{
		eepromdata[EEPROM_OFFSET + i] = convertFromInterface(config.interface[i]);
	}

	if (FT_EE_UAWrite(handle, eepromdata, read) != FT_OK)
	{
		free(eepromdata);
		FT_Close(handle);
		return -5;
	}

	free(eepromdata);
	FT_Close(handle);
	return 0;
}

